/*
 *
 * mpeg_out_driver.c
 * Description:
 * MPEG out driver implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

/*! \file mpeg_out_driver.c
    \brief Implementation of Mpeg out Driver. 
    \code.
*/

#include <linux/module.h>	/* module */
#include <asm/uaccess.h>
#include "pal.h"
#include "soc_interface_driver.h"
#include "mpeg_out_driver.h"

/*****************************************/
/* Definitions for MPEG-Out SoC Unit     */
/*****************************************/

#define PUMA5_PINMUX0_MPEG_OUT_EN            0x4

#define PUMA5_LPSC_MPEG_OUT_MODULE_OFFSET    (0x21A0C)
#define PUMA5_LPSC_MPEG_OUT_MODULE( device ) GET_SOC_REG_ADDR( device, PUMA5_LPSC_MPEG_OUT_MODULE_OFFSET )
#define PUMA5_LPSC_MPEG_OUT_MODULE_DIS_VAL   (0x002)
#define PUMA5_LPSC_MPEG_OUT_MODULE_EN_VAL    (0x103)/* bit 8   Module local reset control
                                                      0 = assert local reset
                                                      1 = de-assert local reset
                                                    bits 0:4 
                                                      00000: SwRstDisable
                                                      00001: SyncRst
                                                      00010: Disable
                                                      00011: Enable */                                                    
/* SoC MPEG Out Registers */
#define PUMA5_MPEG_OUT_RESET_REG_OFFSET          (0x0)
#define PUMA5_MPEG_OUT_CONTROL_REG_OFFSET        (0x4)

#define PUMA5_MPEG_OUT_RESET_REG( device )       GET_MPEG_OUT_REG_ADDR( device, PUMA5_MPEG_OUT_RESET_REG_OFFSET )
#define PUMA5_MPEG_OUT_CONTROL_REG( device )     GET_MPEG_OUT_REG_ADDR( device, PUMA5_MPEG_OUT_CONTROL_REG_OFFSET )

/* Mapping of MPEG Out Reset Registers (PUMA5_MPEG_OUT_RESET_REG). bits 31:5 reserved */
#define PUMA5_MPEG_OUT_PORT1_EN_FIELD           (0x01)      /* bit-0 - value 1 for reset    */
#define PUMA5_MPEG_OUT_PORT2_EN_FIELD           (0x02)      /* bit-1 - value 1 for reset    */
#define PUMA5_MPEG_OUT_PORT3_EN_FIELD           (0x04)      /* bit-2 - value 1 for reset    */
#define PUMA5_MPEG_OUT_PORT4_EN_FIELD           (0x08)      /* bit-3 - value 1 for reset    */
#define PUMA5_MPEG_OUT_S2P_EN_FIELD             (0x10)      /* bit-4 - value 1 for reset    */
#define PUMA5_MPEG_OUT_PORT_EN_VAL              (0) 
#define PUMA5_MPEG_OUT_PORT_DIS_VAL             (1)

/* Mapping of MPEG Out Control Registers (PUMA5_MPEG_OUT_CONTROL_REG). bits 23:12 reserved */
#define PUMA5_MPEG_OUT_CLK_INV_FIELD            (0x100)       /* bit-8 */
#define PUMA5_MPEG_OUT_SYNC_PULSE_FIELD         (0x040)       /* bit-6 */
#define PUMA5_MPEG_OUT_MSB_FIELD                (0x020)       /* bit-5 */
#define PUMA5_MPEG_OUT_SYNC_INV_FIELD           (0x010)       /* bit-4 */
#define PUMA5_MPEG_OUT_VALID_INV_FIELD          (0x008)       /* bit-3 */


#define PUMA5_MPEG_OUT_INTFC_SELECT_OFFSET      (0x000)       /* bits-0:1     */
#define PUMA5_MPEG_OUT_INTFC_SELECT_WIDTH       (0x002)       /* 2 bits width */
#define PUMA5_MPEG_OUT_INTFC_SELECT_DIS_VAL     (0x000)
#define PUMA5_MPEG_OUT_INTFC_SELECT_EN_VAL      (0x001)       /* 00  All bus is zero
                                                                 01  Serial interface
                                                                 10  S-MODE (parallel single mode)
                                                                 11  M-MODE (parallel multiple mode) */

static void mpegOutCloseAllPorts ( socDevice_e device );
                                                                 
static int  mpegOutDriverIoctl   (struct inode *inode,struct file *file, 
                                  unsigned int ioctl_num,unsigned long ioctl_param, struct semaphore *soc_sem);

static SoCDriverOperations_t mpeg_out_operations =  
{
   .socModuleID   = SOC_MPEG_OUT_MODULE_ID,
   .soc_ioctl     = &mpegOutDriverIoctl, 
   .soc_release   = NULL,
   .soc_open      = NULL,
   .soc_cleanup   = NULL,
   .soc_init      = NULL
};

/**************************************************************************/
/*! \fn void mpegOutRegisterSoCModule (SoCDriverOperations_t **mpeg_out_operations)
 **************************************************************************
 *  \brief Register the MPEG-Out driver function pointers object.
 *  \param[in] **mpeg_out_operations - pointer to MPEG-Out .
 *  \return none.
 **************************************************************************/
void mpegOutRegisterSoCModule (SoCDriverOperations_t **mpeg_out_operations_out)
{
   *mpeg_out_operations_out  =  &mpeg_out_operations;
}

/**************************************************************************/
/*! \fn static int   mpegOutDriverIoctl(struct inode *inode, struct file *file,
                                        unsigned int ioctl_num, unsigned long ioctl_param)
 **************************************************************************
 *  \brief This function is called whenever a process tries to do an ioctl to our
 * SoC driver file. We get two extra parameters (additional to the inode and file
 * structures, which all device functions get): the number of the ioctl called
 * and the parameter given to the ioctl function.
 *
 * If the ioctl is write or read/write (meaning output is returned to the
 * calling process), the ioctl call returns the output of this function.
 * _IOC macros are used to decode ioctl numbers:
 *    _IOC_DIR(nr)  - get direction number 
 *    _IOC_TYPE(nr) - get type number 
 *    _IOC_NR(nr)   - get ordinal (sequential) number 
 *    _IOC_SIZE(nr) - get size number 
 *  \return 0 on success otherwise ioctl error code.
 **************************************************************************/
static int   mpegOutDriverIoctl(struct inode *inode,
                                struct file *file,
                                unsigned int ioctl_num,
                                unsigned long ioctl_param,
                                struct semaphore *soc_sem)
{
   /*
    * extract the type and number bitfields, and don't decode
    * wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok(  )
    */
   if (_IOC_TYPE(ioctl_num) != SOC_MPEG_OUT_MODULE_ID) 
      return -ENOTTY;
   if (_IOC_NR(ioctl_num) > SOC_MPEGO_IOCTL_MAXNR) 
      return -ENOTTY;

   /* */
   down(soc_sem);
   /* 
	 * Switch according to the ioctl called 
	 */
	switch (ioctl_num) {
	case SOC_MPEGO_UNIT_ENABLE:
		/* 
		 * This IOCTL command should get out of reset the MPEG_OUT SoC unit. */
      mpegOutUnitEnable( SOC_DEVICE_LOCAL );
	  break;

	case SOC_MPEGO_UNIT_DISABLE:
      /* 
       * This IOCTL command should reset the MPEG_OUT SoC unit. */
      mpegOutUnitDisable( SOC_DEVICE_LOCAL );
      break;

	case SOC_MPEGO_PORT_ENABLE:
		/* 
		 * This IOCTL command should open an MPEG Out port.
		 */
      if ( mpegOutPortSet( (Uint32)ioctl_param, True, SOC_DEVICE_LOCAL ) < 0 )
      {
         up(soc_sem);
         return -EINVAL;
      }
      break;

   case SOC_MPEGO_PORT_DISABLE:
      /* 
		 * This IOCTL command should close an MPEG Out port.
		 * 1. Open MPEG Out ports according to the ioctl_param (port number 1-4). 
		 */
      if ( mpegOutPortSet( (Uint32)ioctl_param, False, SOC_DEVICE_LOCAL ) < 0 )
      {
         up(soc_sem);
         return -EINVAL;
      }
      break;

   case SOC_MPEGO_UNIT_GLOBAL_CONFIG:
   {
      SoCMPEGOutGlobalConfig_t mpegOutConfiguration;
    
      /* 
		 * This IOCTL command should configure the MPEG Out Control Register.
		 * 1. configure the MPEG Out Control Register according to ioctl_param (pointer to SoCMPEGOutGlobalConfig_t). 
		 */
      if ( copy_from_user((void *)(&mpegOutConfiguration),
                          (void *)ioctl_param,
                          sizeof(SoCMPEGOutGlobalConfig_t)) )
      {
         printk(KERN_ERR "mpegOutDriverIoctl:: Invalid argument for MPEG Out global config.\n");
         up(soc_sem);
         return -EINVAL;
      }
      
      mpegOutUnitGlobalConfig( &mpegOutConfiguration, SOC_DEVICE_LOCAL );
   }
   break;

   default:  /* redundant, as cmd was checked against MAXNR */
      up(soc_sem);
      return -ENOTTY;
	}

   up(soc_sem);
	return 0;
}

static void mpegOutCloseAllPorts( socDevice_e device )
{
   Uint32 regValue;

   /* Close all MPEG Out ports */
   printk(KERN_INFO "mpegOutCloseAllPorts:: Close all MPEG Out ports.\n");

   socReadRegister((PUMA5_MPEG_OUT_RESET_REG( device )),&regValue,True);
   /* Writing value of 1 will reset the MPEG ports */
   regValue |= (PUMA5_MPEG_OUT_PORT1_EN_FIELD |  /* reset port 1 */
                PUMA5_MPEG_OUT_PORT2_EN_FIELD |  /* reset port 2 */
                PUMA5_MPEG_OUT_PORT3_EN_FIELD |  /* reset port 3 */
                PUMA5_MPEG_OUT_PORT4_EN_FIELD |  /* reset port 4 */
                PUMA5_MPEG_OUT_S2P_EN_FIELD);    /* reset s2p    */
   socWriteRegister((PUMA5_MPEG_OUT_RESET_REG( device )),regValue,True);
}

Int32 mpegOutUnitGlobalConfig( SoCMPEGOutGlobalConfig_t *mpegOutConfiguration, socDevice_e device )
{
      Uint32 regValue;

      printk(KERN_INFO "mpegOutUnitGlobalConfig:: Configure the MPEG Out Control Registers.\n");
      /* 1. configure the MPEG Out Control Register according to ioctl_param (pointer to SoCMPEGOutGlobalConfig_t) */
      socReadRegister((PUMA5_MPEG_OUT_CONTROL_REG( device )),&regValue,True);
      if (mpegOutConfiguration->clockInv == 1)
         regValue |= (PUMA5_MPEG_OUT_CLK_INV_FIELD);  /* set 1 */
      else
         regValue &= ~(PUMA5_MPEG_OUT_CLK_INV_FIELD); /* set 0 */

      if (mpegOutConfiguration->msbFirstTxOrder == 1)
         regValue |= (PUMA5_MPEG_OUT_MSB_FIELD);
      else
         regValue &= ~(PUMA5_MPEG_OUT_MSB_FIELD);

      if (mpegOutConfiguration->syncInv == 1)
         regValue |= (PUMA5_MPEG_OUT_SYNC_INV_FIELD);
      else
         regValue &= ~(PUMA5_MPEG_OUT_SYNC_INV_FIELD);

      if (mpegOutConfiguration->syncPulseWidth == 1)
         regValue |= (PUMA5_MPEG_OUT_SYNC_PULSE_FIELD);
      else
         regValue &= ~(PUMA5_MPEG_OUT_SYNC_PULSE_FIELD);

      if (mpegOutConfiguration->validInv == 1)
         regValue |= (PUMA5_MPEG_OUT_SYNC_INV_FIELD);
      else
         regValue &= ~(PUMA5_MPEG_OUT_SYNC_INV_FIELD);

      socWriteRegister((PUMA5_MPEG_OUT_CONTROL_REG( device )),regValue,True);
      
      return 0;
}

Int32 mpegOutPortSet( Uint32 port, Bool enable, socDevice_e device )
{
      Uint32 regValue;
      Uint32 data;

      /* 1. Open/Close MPEG Out ports according to the ioctl_param (port number 1-4). */
                
      if (port < 1 || port > 4)
      {
         printk(KERN_ERR "mpegOutPortSet:: Invalid argument (MPEG Out port %d). \n",port);
         return -EINVAL;
      }

      if ( enable == False )
      {
         data = PUMA5_MPEG_OUT_PORT_DIS_VAL;
      }
      else
      {
         data = PUMA5_MPEG_OUT_PORT_EN_VAL;
      }
      
      /* 1. Close MPEG Out ports according to the ioctl_param (port number 1-4) */
      printk(KERN_INFO "mpegOutPortSet:: %s MPEG Out port %d. \n", enable == False ? "Close" : "Open", port);
      socReadRegister((PUMA5_MPEG_OUT_RESET_REG( device )),&regValue,True);
      /* Reset signals are active high (0-port enable 1-port disable .*/
      socBitField32Set(&regValue,data,(port-1),1);
      socWriteRegister((PUMA5_MPEG_OUT_RESET_REG( device )),regValue,True);
      
      return 0;
}

Int32 mpegOutUnitEnable( socDevice_e device )
{
      Uint32 regValue;

	   /* 1. Kick the bootCfg lock registers.
	    * 2. Write 1 to the "MPEG Out EN" (bit 2) to enable MPEG Out unit.
        * 3. Set LPSC_MPEG_OUT_MODULE state to Enable.
        * 4. Close all MPEG Out ports.
        * 5. Enable MPEG-Out serial interface. 
		 */
      /* 1. Kick the bootCfg lock registers */
      printk(KERN_INFO "mpegOutUnitEnable:: Kick the BootCfg lock. \n");
      socWriteRegister((PUMA5_BOOTCFG_KICK_0_REG( device )), PUMA5_BOOTCFG_KICK_0_VAL, True);
      socWriteRegister((PUMA5_BOOTCFG_KICK_1_REG( device )), PUMA5_BOOTCFG_KICK_1_VAL, True);

      /* 2. Bit 2 "MPEG Out EN" When active, MPEG Out is ported out over p-flash. */
      printk(KERN_INFO "mpegOutUnitEnable:: Enable MPEG Out unit. \n");
      socReadRegister((PUMA5_PINMUX0_REG( device )),&regValue,True);
      regValue |= PUMA5_PINMUX0_MPEG_OUT_EN;
      socWriteRegister((PUMA5_PINMUX0_REG( device )), regValue, True);

      /* 3. Set LPSC_MPEG_OUT_MODULE state to Enable */
      socWriteRegister((PUMA5_LPSC_MPEG_OUT_MODULE( device )), PUMA5_LPSC_MPEG_OUT_MODULE_EN_VAL,True);
      socWriteRegister((PUMA5_PSC_GO_REG( device )), PUMA5_PSC_GO_VAL  ,True); 

      /* 4. Close all MPEG Out ports */
      mpegOutCloseAllPorts( device );

      /* 5. Enable MPEG-Out serial interface */
      socReadRegister((PUMA5_MPEG_OUT_CONTROL_REG( device )),&regValue,True);
      socBitField32Set(&regValue,
                       PUMA5_MPEG_OUT_INTFC_SELECT_EN_VAL,
                       PUMA5_MPEG_OUT_INTFC_SELECT_OFFSET,
                       PUMA5_MPEG_OUT_INTFC_SELECT_WIDTH);
      socWriteRegister((PUMA5_MPEG_OUT_CONTROL_REG( device )),regValue,True);

      printk(KERN_INFO "mpegOutUnitEnable:: MPEG Out is working. \n");
      /* MPEG Out Control Register is now with the default values */

      return 0;
}

Int32 mpegOutUnitDisable( socDevice_e device )
{
      Uint32 regValue;

      /* 
       * 1. Kick the bootCfg lock registers.
       * 2. Close all MPEG Out ports. 
       * 3. Disable MPEG-Out serial interface (all bus is zero). 
       * 4. Write 0 to the "MPEG Out EN" (bit 2) to disable MPEG Out unit.
       * 5. Set LPSC_MPEG_OUT_MODULE state to Disable. 
       */

      /* 1. Kick the bootCfg lock registers */
	  printk(KERN_INFO "mpegOutUnitDisable:: Kick the BootCfg lock. \n");
      socWriteRegister((PUMA5_BOOTCFG_KICK_0_REG( device )), PUMA5_BOOTCFG_KICK_0_VAL, True);
      socWriteRegister((PUMA5_BOOTCFG_KICK_1_REG( device )), PUMA5_BOOTCFG_KICK_1_VAL, True);

      /* 2. Close all MPEG Out ports */
      mpegOutCloseAllPorts( device );

      /* 3. Disable MPEG-Out serial interface (all bus is zero) */
      socReadRegister((PUMA5_MPEG_OUT_CONTROL_REG( device )),&regValue,True);
      socBitField32Set(&regValue,
                       PUMA5_MPEG_OUT_INTFC_SELECT_DIS_VAL,
                       PUMA5_MPEG_OUT_INTFC_SELECT_OFFSET,
                       PUMA5_MPEG_OUT_INTFC_SELECT_WIDTH);
      socWriteRegister((PUMA5_MPEG_OUT_CONTROL_REG( device )),regValue,True);

      /* 4. Bit 2 "MPEG Out EN" When active, MPEG Out is ported out over p-flash. */
      printk(KERN_INFO "mpegOutUnitDisable:: Disable MPEG Out unit. \n");
      socReadRegister((PUMA5_PINMUX0_REG( device )),&regValue,True);
      regValue |= ~(PUMA5_PINMUX0_MPEG_OUT_EN);
      socWriteRegister((PUMA5_PINMUX0_REG( device )), regValue, True);

      /* 5. Set LPSC_MPEG_OUT_MODULE state to Disable */
      socWriteRegister(PUMA5_LPSC_MPEG_OUT_MODULE( device ), PUMA5_LPSC_MPEG_OUT_MODULE_DIS_VAL,True);
      socWriteRegister((PUMA5_PSC_GO_REG( device )), PUMA5_PSC_GO_VAL  ,True); 

      return 0;
}

EXPORT_SYMBOL( mpegOutUnitGlobalConfig );
EXPORT_SYMBOL( mpegOutPortSet );
EXPORT_SYMBOL( mpegOutUnitEnable );
EXPORT_SYMBOL( mpegOutUnitDisable );
